/**
 *  @file   GrowingWindow.h
 *  @brief  GrowingWindow class C wrapper header
 *  @date   17.09.2013
 *  @author Peter Rudolph
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES\n
 *  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF\n
 *  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR\n
 *  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES\n
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN\n
 *  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF\n
 *  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.\n
 *
 *  @copyright Intelligent Embedded Systems, University of Kassel, 2013. All rights reserved.
 *  @license This library is released under <a href="http://www.gnu.org/licenses/lgpl-3.0.de.html">LGPLv3</a>.
 */

#ifndef __GROWINGWINDOW_H__
#define __GROWINGWINDOW_H__

#ifdef __cplusplus
extern "C" {
#endif

/**
 *  @brief Struct representing GrowingWindow handle
 */
typedef struct GrowingWindow GrowingWindow;

/**
 *  @brief Create GrowingWindow Handle
 *  @param degree Degree of approximated polynomial
 *  @return Pointer to GrowingWindow Handle
 */
GrowingWindow* GrowingWindow_create(int degree);

/**
 *  @brief Reset GrowingWindow Approximation
 *  @param gw Pointer to GrowingWindow Handle
 */
void GrowingWindow_reset(GrowingWindow* gw);

/**
 *  @brief Update GrowingWindow Approximation
 *  @param gw Pointer to GrowingWindow Handle
 */
int GrowingWindow_getNumSamples(GrowingWindow* gw);

/**
 *  @brief Get number of samples used by GrowingWindow
 *  @param gw Pointer to GrowingWindow Handle
 *  @param y Value to update with.
 *  @return Number of Samples
 */
void GrowingWindow_update(GrowingWindow* gw, double y);

/**
*	@brief Evaluate polynomial at x.
*	@param gw Pointer to GrowingWindow handle.
*	@param x Value polynomial is evaluated at.
*	@return Evaluation at x.
*/
double GrowingWindow_eval(GrowingWindow* gw, double x);

/**
 *  @brief Get orthogonal coefficients
 *  @param gw Pointer to GrowingWindow Handle
 *  @param coefficients Pointer to double array which is filled by function itself, size must be degree + 1
 */
void GrowingWindow_getOrthogonalCoefficients(GrowingWindow* gw, double *coefficients);

/**
*	@brief Get approximation error.
*	@param gw Pointer to GrowingWindow handle
*	@return Approximation error
*/
double GrowingWindow_getApproximationError(GrowingWindow* gw);

/**
*	@brief Get extremum position.
*	@param gw Pointer to GrowingWindow handle
*	@return Extremum position
*/
double GrowingWindow_getExtremumPosition(GrowingWindow* gw);

/**
 *  @brief Delete GrowingWindow Handle
 *  @param gw Pointer to GrowingWindow Handle
 */
void GrowingWindow_delete(GrowingWindow* gw);

#ifdef __cplusplus
}
#endif

#endif /*__GROWINGWINDOW_H__*/
