/**
 *  @file   OrthogonalPolynomialSlidingWindow.h
 *  @brief  OrthogonalPolynomialSlidingWindow class C wrapper header
 *  @date   17.09.2013
 *  @author Peter Rudolph
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES\n
 *  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF\n
 *  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR\n
 *  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES\n
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN\n
 *  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF\n
 *  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.\n
 *
 *  @copyright Intelligent Embedded Systems, University of Kassel, 2013. All rights reserved.
 *  @license This library is released under <a href="http://www.gnu.org/licenses/lgpl-3.0.de.html">LGPLv3</a>.
 */

#ifndef __ORTHOGONALPOLYNOMIALSLIDINGWINDOW_H__
#define __ORTHOGONALPOLYNOMIALSLIDINGWINDOW_H__

#ifdef __cplusplus
extern "C" {
#endif

/**
 *  @brief Struct representing OrthogonalPolynomialSlidingWindow handle
 */
typedef struct OrthogonalPolynomialSlidingWindow OrthogonalPolynomialSlidingWindow;

/**
 *  @brief Create OrthogonalPolynomialSlidingWindow Handle
 *  @param degree Degree of approximated polynomial
 *  @param windowsize Size of sliding window
 *  @return Pointer to OrthogonalPolynomialSlidingWindow Handle
 */
OrthogonalPolynomialSlidingWindow* OrthogonalPolynomialSlidingWindow_create(int degree, int windowsize);

/**
 *  @brief Restart OrthogonalPolynomialSlidingWindow Approximation
 *  @param sw Pointer to OrthogonalPolynomialSlidingWindow Handle
 */
void OrthogonalPolynomialSlidingWindow_restart(OrthogonalPolynomialSlidingWindow* sw);

/**
 *  @brief Update OrthogonalPolynomialSlidingWindow Approximation
 *  @param sw Pointer to OrthogonalPolynomialSlidingWindow Handle
 *  @param y Value to update with.
 */
void OrthogonalPolynomialSlidingWindow_update(OrthogonalPolynomialSlidingWindow* sw, double y);

/**
*	@brief Evaluate polynomial at x.
*	@param sw Pointer to OrthogonalPolynomialSlidingWindow handle.
*	@param x Value polynomial is evaluated at.
*	@return Evaluation at x.
*/
double OrthogonalPolynomialSlidingWindow_eval(OrthogonalPolynomialSlidingWindow* sw, double x);

/**
 *  @brief Get coefficients
 *  @param sw Pointer to OrthogonalPolynomialSlidingWindow Handle
 *  @param coefficients Pointer to double array which is filled by function itself, size must be degree + 1
 */
void OrthogonalPolynomialSlidingWindow_getOrthogonalCoefficients(OrthogonalPolynomialSlidingWindow* sw, double *coefficients);

/**
*	@brief Get approximation error.
*	@param sw Pointer to OrthogonalPolynomialSlidingWindow handle
*	@return Approximation error
*/
double OrthogonalPolynomialSlidingWindow_getApproximationError(OrthogonalPolynomialSlidingWindow* sw);

/**
*	@brief Get extremum position.
*	@param sw Pointer to OrthogonalPolynomialSlidingWindow handle
*	@return Extremum position
*/
double OrthogonalPolynomialSlidingWindow_getExtremumPosition(OrthogonalPolynomialSlidingWindow* sw);

/**
 *  @brief Delete OrthogonalPolynomialSlidingWindow Handle
 *  @param sw Pointer to OrthogonalPolynomialSlidingWindow Handle
 */
void OrthogonalPolynomialSlidingWindow_delete(OrthogonalPolynomialSlidingWindow* sw);

#ifdef __cplusplus
}
#endif

#endif /*__ORTHOGONALPOLYNOMIALSLIDINGWINDOW_H__*/