/**
 *  @file   TrigonometricPolynomialSlidingWindow.hpp
 *  @brief  TrigonometricPolynomialSlidingWindow class header
 *  @date   26.03.2013
 *  @author Dr. Thiemo Gruber, Peter Rudolph
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES\n
 *  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF\n
 *  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR\n
 *  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES\n
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN\n
 *  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF\n
 *  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.\n
 *
 *  @copyright Intelligent Embedded Systems, University of Kassel, 2013. All rights reserved.
 *  @license This library is released under <a href="http://www.gnu.org/licenses/lgpl-3.0.de.html">LGPLv3</a>.
 */

#ifndef __TRIGONOMETRICPOLYNOMIALSLIDINGWINDOW_HPP__
#define __TRIGONOMETRICPOLYNOMIALSLIDINGWINDOW_HPP__

#include <vector>

/**
 *  @brief This class implements an updatable sliding window least squares polynomial
 *         approximation using trigonometric polynomials.
 *
 *  @see   "Erich Fuchs, Tobias Hanning and Oliver Schwarz, 'An Update Algorithm
 *         for Fourier Coefficients', 12th European Signal Processing Conference,
 *         pp. 1509-1512, Vienna, Austria, 2004"
 */
class TrigonometricPolynomialSlidingWindow {

public:
  /**
   * @brief Constructor
   *
   * @param degree Degree of polynomial approximation
   * @param windowSize Size of sliding window
   */
	TrigonometricPolynomialSlidingWindow(int degree, int windowSize);

	/**
	 * @brief Initialize sliding window.
	 */
	void init();

	/**
	 * @brief Restart sliding window.
	 */
	void restart();

	/**
	 * @brief Update sliding window approximation.
	 * @param y Value to update with.
	 */
	void update(double y);

	/**
	 * @brief Evaluate polynomial at x.
	 * @param x Value to evaluate polynomial at.
	 * @return Evaluation
	 */
	double eval(double x);

	/**
	 * @brief Get coefficients.
	 * @return Vectors of coefficients
	 */
	std::vector< std::vector<double> > getCoefficients();

  /**
   * @brief Get approximation error.
   * @return Approximation error
   */
	double getApproximationError();

private:
	int degree; /**< degree */
	int windowSize; /**< window size */
        std::vector<double> cRe; /**< vector real */
        std::vector<double> cIm; /**< vector imaginary */
        std::vector<double> cosPhi; /**< vector cos phi */
	std::vector<double> sinPhi; /**< vector sin phi */
        std::vector<double> values; /**< vector values */
	int valueIdx; /**< value idx  */
	double sqsum; /**< square sum */
};

#endif /* #ifndef __TRIGONOMETRICPOLYNOMIALSLIDINGWINDOW_HPP__ */

