/**
 * 	This class provides algorithms for up- and downdating least squares
 * 	polynomial fits to discrete data with orthogonal polynomials using
 * 	(hyper-)rotations and for evaluating the resulting polynomials.
 * 
 *  @see   Fuchs, E.; Gruber, T.; Nitschke, J.; Sick, B.,
 *			"Online Segmentation of Time Series Based on Polynomial Least-Squares Approximations,"
 * 			Pattern Analysis and Machine Intelligence, IEEE Transactions on ,
 * 			vol.32, no.12, pp.2232,2245, Dec. 2010
 *
 *  @date   26.03.2013
 *  @author Dr. Thiemo Gruber, Peter Rudolph
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES\n
 *  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF\n
 *  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR\n
 *  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES\n
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN\n
 *  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF\n
 *  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.\n
 *
 *  @copyright Intelligent Embedded Systems, University of Kassel, 2013. All rights reserved.
 *  @license This library is released under <a href="http://www.gnu.org/licenses/lgpl-3.0.de.html">LGPLv3</a>.
 */
package libffal;

public class GrowingWindow {

    private double[] alpha, beta;
    private double[] d;
    private double sigma;
    private int n;
    private int degree;
    private double ysqsum;
    private int nbPts;

    /**
     * @param degree Degree of polynomial approximation 
     */
    public GrowingWindow(int degree) {

        if (degree < 0)
            throw new IllegalArgumentException("Degree must not be negative.");

        alpha = new double[degree + 2];
        beta = new double[degree + 1];
        d = new double[degree + 2];
        this.degree = degree;
        sigma = 0;
        n = 0;
        nbPts = 0;
        ysqsum = 0;
    }

    /**
     * Reset GrowingWindow
     */
    public void reset() {
        n = 0;
        nbPts = 0;
    }

    /**
     * Update polynomial approximation at y.
     * 
     * @param y Value to update with.
     */
    public void update(double y) {

        if (n == 0) {
            sigma = 1;
            alpha[0] = 0;
            d[0] = y;
            n = 1;
            nbPts = 1;
            ysqsum = y * y;

        } else {

            int nNew = n + (n < degree + 2 ? 1 : 0);
            double sigmaNew = Math.sqrt(sigma * sigma + 1);
            double c = sigma / sigmaNew;
            double s = 1 / sigmaNew;
            sigma = sigmaNew;
            double theta1 = c * s * (nbPts - alpha[0]);
            double theta2 = degree > 0 ? -s * beta[0] : 0;
            alpha[nNew - 1] = c * c * nbPts + s * s * alpha[0];
            alpha[0] = c * c * alpha[0] + s * s * nbPts;
            beta[0] = c * beta[0];
            d[nNew - 1] = c * y - s * d[0];
            d[0] = c * d[0] + s * y;

            for (int i = 1; i < nNew - 1; i++) {
                if (beta[i - 1] * beta[i - 1] + theta1 * theta1 == 0) {
                    n = i;
                    return;
                }

                double zeta = Math.sqrt(beta[i - 1] * beta[i - 1] + theta1
                        * theta1);
                c = beta[i - 1] / zeta;
                s = theta1 / zeta;
                beta[i - 1] = zeta;
                double t = c * c * alpha[i] + 2 * c * s * theta2 + s * s
                        * alpha[nNew - 1];
                double r = c * c * alpha[nNew - 1] - 2 * c * s * theta2 + s * s
                        * alpha[i];
                theta1 = c * s * (alpha[nNew - 1] - alpha[i]) + theta2
                        * (c * c - s * s);
                theta2 = -s * beta[i];
                alpha[i] = t;
                alpha[nNew - 1] = r;
                beta[i] = c * beta[i];
                t = d[i];
                d[i] = c * t + s * d[nNew - 1];
                d[nNew - 1] = c * d[nNew - 1] - s * t;
            }

            if (theta1 != 0) {
                beta[nNew - 2] = Math.abs(theta1);
                d[nNew - 1] = Math.signum(theta1) * d[nNew - 1];
                n = nNew;
            } else {
                System.err.println("Warning: theta1 = 0");
            }

            nbPts++;
            ysqsum += y * y;

        }
    }

    /**
     * Evaluate polynomial approximation at x.
     * 
     * @param x Value to evaluate polynomial at.
     */
    public double eval(double x) {

        if (n <= degree)
            throw new IllegalStateException("Unable to evaluate polynomial for"
                    + " less than degree + 1 points.");

        double[] y = new double[d.length];

        if (degree == 0) {
            y[0] = d[0];

        } else if (degree == 1) {
            y[0] = d[0] + d[1] / beta[0] * (x - alpha[0]);

        } else {
            y[degree] = d[degree] / beta[degree - 1];
            y[degree - 1] = (d[degree - 1] + y[degree]
                    * (x - alpha[degree - 1]))
                    / beta[degree - 2];

            for (int i = degree - 2; i >= 1; i--) {
                y[i] = (d[i] + y[i + 1]
                        * (x - alpha[i]) - y[i + 2]
                        * beta[i])
                        / beta[i - 1];
            }

            y[0] = d[0] + y[1] * (x - alpha[0]) - y[2] * beta[0];
        }

        return y[0] / sigma;
    }

    public int getNumSamples() {
        return nbPts;
    }

    /**
     * Get orthogonal coefficients.
     */
    public double[] getOrthogonalCoefficients() {

        if (n <= degree)
            throw new IllegalStateException("Unable to evaluate polynomial for"
                    + " less than degree + 1 points.");

        double[] c = new double[degree + 1];
        double nf = 1 / sigma;

        c[0] = nf * d[0];

        for (int i = 1; i <= degree; i++) {
            nf /= beta[i - 1];
            c[i] = d[i] * nf;
        }

        return c;
    }

    /**
     * Get approximation error.
     */
    public double getApproximationError() {

        if (n <= degree)
            throw new IllegalStateException("Unable to evaluate "
                    + "approximation error for less than degree + 1 points.");

        double dsqsum = 0;

        for (int i = 0; i <= degree; i++)
            dsqsum += d[i] * d[i];

        if (ysqsum - dsqsum < -1e-8)
            System.err.println("Warning: Approximation error negative!");
        
        return Math.max(0, ysqsum - dsqsum);

    }

    /**
     * Get extremum position
     */
    public int getExtremumPosition() {

        if (degree < 2 || n <= degree)
            throw new IllegalStateException("Unable to calculate the extremum "
                    + "of the current solution (degree must be >= 2, solution "
                    + "size must be > degree)");

        if (d[2] == 0)
            throw new IllegalStateException("Unable to calculate the extremum "
                    + "of the current solution (curve is zero)");

        return (int) Math.round((alpha[0] + alpha[1]) / 2 - d[1] * beta[1]
                / (2 * d[2]));
    }
}
