/**
 *  This class implements an updatable sliding window least squares polynomial
 *  approximation using discrete Legendre polynomials.
 * 
 * 	@see "Erich Fuchs, 'Schnelle Quadratmittelapproximation in gleitenden
 *        Zeitfenstern mit diskreten orthogonalen Polynomen', Ph.D. thesis,
 *        Universitaet Passau, Fakultaet fuer Mathematik und Informatik, 1999"
 * 
 *  @date   26.03.2013
 *  @author Dr. Thiemo Gruber, Peter Rudolph
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES\n
 *  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF\n
 *  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR\n
 *  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES\n
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN\n
 *  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF\n
 *  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.\n
 *
 *  @copyright Intelligent Embedded Systems, University of Kassel, 2013. All rights reserved.
 *  @license This library is released under <a href="http://www.gnu.org/licenses/lgpl-3.0.de.html">LGPLv3</a>.
 */

package libffal;

public class OrthogonalPolynomialSlidingWindow {

    private int degree;
    private int windowSize;
    private double[] qnorm;
    private double[] delta;
    private double[] alpha;
    private double[][] updateCoefs;
    private double[] values;
    private int valueIdx;
    private double sqsum;

    /**
     * @param degree Degree of polynomial approximation 
     * @param windowSize Size of sliding window
     */
    public OrthogonalPolynomialSlidingWindow(int degree, int windowSize) {
        if (degree < 0)
            throw new IllegalArgumentException("Degree must not be negative.");

        if (degree >= windowSize)
            throw new IllegalArgumentException(
                    "Window size must be greater than degree.");

        this.degree = degree;
        this.windowSize = windowSize;

        init();

        double m = windowSize - 1;

        double[] delta = new double[degree + 2];
        double[] qnorm = new double[degree + 2];

        delta[0] = m + 1;
        qnorm[0] = delta[0];

        double qmpo = (m + 1) * (m + 1);
        for (int i = 1; i <= degree + 1; i++) {
            double qi = i * i;
            delta[i] = qi * (qmpo - qi) / (16 * qi - 4);
            qnorm[i] = qnorm[i - 1] * delta[i];
        }

        this.delta = new double[degree + 1];
        System.arraycopy(delta, 0, this.delta, 0, degree + 1);
        this.qnorm = new double[degree + 1];
        System.arraycopy(qnorm, 0, this.qnorm, 0, degree + 1);

        double[][] gamma = new double[degree + 1][];
        gamma[0] = new double[] { 1, 0, 0 };
        if (degree > 0) {
            gamma[1] = new double[] { 1, 1, 0, 0 };
            for (int i = 2; i < gamma.length; i++) {
                gamma[i] = new double[i + 3];

                for (int j = 0; j <= i; j++) {
                    gamma[i][j] = gamma[i - 1][j + 1] * delta[j + 1]
                            + gamma[i - 1][j] - gamma[i - 2][j] * delta[i - 1];
                    if (j > 0)
                        gamma[i][j] += gamma[i - 1][j - 1];
                }
            }
        }

        updateCoefs = new double[degree + 1][];
        updateCoefs[0] = new double[2];
        updateCoefs[0][0] = 1;
        updateCoefs[0][1] = -1;

        if (degree > 0) {
            for (int i = 1; i < updateCoefs.length; i++) {
                updateCoefs[i] = new double[i + 2];
                updateCoefs[i][0] = updateCoefs[i - 1][0] * i / (4.0 * i - 2)
                        * (m + i + 1);
                updateCoefs[i][1] = updateCoefs[i - 1][1] * -i / (4.0 * i - 2)
                        * (m - i + 1);
                for (int j = 2; j < i + 2; j++)
                    updateCoefs[i][j] = -gamma[i][j - 2];
            }
        }
    }

    /**
     * Initialize sliding window approximation.
     */
    private void init() {
        this.alpha = new double[degree + 1];
        this.values = new double[windowSize];
        this.valueIdx = 0;
        this.sqsum = 0;
    }

    /**
     * Restart sliding window approximation.
     */
    public void restart() {
        double[] valuesOld = values;
        int valueIdxOld = valueIdx;

        init();

        for (int i = 0; i < windowSize; i++)
            update(valuesOld[(valueIdxOld + i) % valuesOld.length]);
    }

    /**
     * Update polynomial approximation at y.
     * 
     * @param y Value to update with.
     */
    public void update(double y) {

        for (int i = 0; i <= degree; i++) {
            double alphaNew = alpha[i] + y * updateCoefs[i][0]
                    + values[valueIdx] * updateCoefs[i][1];

            for (int j = 0; j < i; j++)
                alphaNew += updateCoefs[i][j + 2] * alpha[j];

            alpha[i] = alphaNew;
        }
        sqsum += y * y - values[valueIdx] * values[valueIdx];
        values[valueIdx++] = y;
        if (valueIdx >= values.length)
            valueIdx = 0;
    }

    /**
     * Evaluate polynomial approximation at x.
     * 
     * @param x Value to evaluate polynomial at.
     */
    public double eval(double x) {

        double[] q = new double[degree + 1];
        q[degree] = alpha[degree] / qnorm[degree];

        if (degree > 0) {
            x = x - (windowSize - 1) / 2.0;
            q[degree - 1] = alpha[degree - 1] / qnorm[degree - 1] + x
                    * q[degree];

            for (int i = degree - 2; i >= 0; i--)
                q[i] = alpha[i] / qnorm[i] + x * q[i + 1] - delta[i + 1]
                        * q[i + 2];
        }
        return q[0];
    }

    /**
     * Get orthogonal coefficients.
     */
    public double[] getOrthogonalCoefficients() {

        double[] coefs = new double[degree + 1];

        for (int i = 0; i < coefs.length; i++)
            coefs[i] = alpha[i] / qnorm[i];

        return coefs;
    }

    /**
     * Get approximation error.
     */
    public double getApproximationError() {

        double tmpSum = sqsum;

        for (int i = 0; i <= degree; i++)
            tmpSum -= alpha[i] * alpha[i] / qnorm[i];

        if (tmpSum < -1e-8)
            System.err.println("Warning: Approximation error negative!");

        return Math.max(0, tmpSum);
    }

    /**
     * Get extremum position
     */
    public int getExtremumPosition() {

        if (degree < 2)
            throw new IllegalStateException("Unable to calculate the extremum "
                    + "of the current solution (degree must be >= 2)");

        if (alpha[2] == 0)
            throw new IllegalStateException("Unable to calculate the extremum "
                    + "of the current solution (curve is zero)");

        return (int) Math.round((windowSize - 1) / 2.0 - (alpha[1] * qnorm[2])
                / (2.0 * alpha[2] * qnorm[1]));
    }
}
