/**
 * This class implements an updatable sliding window least squares polynomial
 * approximation using trigonometric polynomials.
 *
 * @see "Erich Fuchs, Tobias Hanning and Oliver Schwarz, 'An Update Algorithm
 * for Fourier Coefficients', 12th European Signal Processing Conference,
 * pp. 1509-1512, Vienna, Austria, 2004"
 * 
 * @author Dr. Thiemo Gruber
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES\n
 *  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF\n
 *  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR\n
 *  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES\n
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN\n
 *  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF\n
 *  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.\n
 *
 *  @copyright Intelligent Embedded Systems, University of Kassel, 2013. All rights reserved.
 *  @license This library is released under <a href="http://www.gnu.org/licenses/lgpl-3.0.de.html">LGPLv3</a>.
 */
package libffal;

public class TrigonometricPolynomialSlidingWindow {

    private int degree;
    private int windowSize;

    private double[] cRe, cIm;
    private double[] cosPhi, sinPhi;
    private double[] values;

    private int valueIdx;
    private double sqsum;

    /**
     * @param degree Degree of polynomial approximation 
     * @param windowSize Size of sliding window
     */
    public TrigonometricPolynomialSlidingWindow(int degree, int windowSize) {

        if (degree < 0) throw new IllegalArgumentException(
            "Degree must not be negative.");

        if (windowSize < 1 || windowSize < degree * 2)
            throw new IllegalArgumentException("Window size must be greater "
                + "than or equal to degree * 2 and must be at least 1.");

        this.degree = degree;
        this.windowSize = windowSize;

        cosPhi = new double[degree + 1];
        sinPhi = new double[degree + 1];

        for (int k = 0; k <= degree; k++) {
            double phi = 2 * Math.PI * k / windowSize;
            cosPhi[k] = Math.cos(phi);
            sinPhi[k] = Math.sin(phi);
        }

        init();
    }

    /**
     * Initialize sliding window approximation.
     */
    private void init() {
        cRe = new double[degree + 1];
        cIm = new double[degree + 1];
        values = new double[windowSize];
        valueIdx = 0;
        sqsum = 0;
    }

    /**
     * Restart sliding window approximation.
     */
    public void reset() {
        init();
    }

    /**
     * Update polynomial approximation at y.
     * 
     * @param y Value to update with.
     */
    public void update(double y) {

        for (int k = 0; k < cRe.length; k++) {
            double s = cRe[k] + y - values[valueIdx];
            cRe[k] = cosPhi[k] * s - sinPhi[k] * cIm[k];
            cIm[k] = cosPhi[k] * cIm[k] + sinPhi[k] * s;
        }

        sqsum += y * y - values[valueIdx] * values[valueIdx];

        values[valueIdx++] = y;

        if (valueIdx >= values.length)
            valueIdx = 0;
    }

    /**
     * Evaluate polynomial approximation at x.
     * 
     * @param x Value to evaluate polynomial at.
     */
    public double eval(double x) {

        double p = cRe[0];

        for (int k = 1; k <= degree - 2 * degree / windowSize; k++) {
            double komegax = k * 2 * Math.PI / values.length * x;
            p += 2 * cRe[k] * Math.cos(komegax)
                 - 2 * cIm[k] * Math.sin(komegax);
        }

        if (2 * degree == windowSize) {
            p += cRe[cRe.length - 1] * Math.cos(Math.PI * x);
        }

        return p / values.length;
    }

    /**
     * Get coefficients.
     */
    public double[][] getCoefficients() {
        double[][] c = new double[2][];
        c[0] = cRe.clone();
        c[1] = cIm.clone();
        return c;
    }

    /**
     * Get approximation error.
     */
    public double getApproximationError() {

        double[][] qnorms = new double[2][degree + 1];
        for (int x = 0; x < windowSize; x++) {
            for (int k = 0; k < cRe.length; k++) {
                double komegax = k * 2 * Math.PI / windowSize * x;
                qnorms[0][k] += Math.pow(Math.cos(komegax), 2);
                qnorms[1][k] += Math.pow(Math.sin(komegax), 2);
            }
        }

        double tmpSum = sqsum;

        for (int i = 0; i <= degree; i++) {
            tmpSum -= Math.pow(cRe[i], 2) / qnorms[0][i];
        }

        for (int i = 1; i <= degree - 2 * degree / windowSize; i++) {
            tmpSum -= Math.pow(cIm[i], 2) / qnorms[1][i];
        }

        if (tmpSum < -1e-8)
            System.err.println("Warning: Approximation error negative!" + tmpSum);

        return Math.max(0, tmpSum);
    }



}

