% /**
%  * 	This class provides algorithms for up- and downdating least squares
%  * 	polynomial fits to discrete data with orthogonal polynomials using
%  * 	(hyper-)rotations and for evaluating the resulting polynomials.
%  * 
%  *  @see   Fuchs, E.; Gruber, T.; Nitschke, J.; Sick, B.,
%  *			"Online Segmentation of Time Series Based on Polynomial Least-Squares Approximations,"
%  * 			Pattern Analysis and Machine Intelligence, IEEE Transactions on ,
%  * 			vol.32, no.12, pp.2232,2245, Dec. 2010
%  *
%  *  @date   16.08.2013
%  *  @author Dr. Thiemo Gruber, Andr Gensler
%  *
%  *  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES\n
%  *  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF\n
%  *  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR\n
%  *  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES\n
%  *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN\n
%  *  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF\n
%  *  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.\n
%  *
%  *  @copyright Intelligent Embedded Systems, University of Kassel, 2013. All rights reserved.
%  *  @license This library is released under <a href="http://www.gnu.org/licenses/lgpl-3.0.de.html">LGPLv3</a>.
%  */

classdef GrowingWindow < handle & matlab.mixin.Copyable
    %GrowingWindow This is the GrowingWindow implementation in Matlab.
    %   For detailed documentation see http://ies-research.de/Software
    
    properties
        alpha
        beta
        d
        sigma
        n
        degree
        ysqsum
        nbPts
    end
    
    methods
        function GW = GrowingWindow(degree)
           if degree <0
                error('Degree must not be negative.')
           end
           GW.alpha = zeros(degree+2,1);
           GW.beta = zeros(degree+1,1);
           GW.d = zeros(degree+2,1);
           GW.degree = degree;
           GW.sigma = 0;
           GW.n = 0;
           GW.nbPts = 0;
           GW.ysqsum = 0;
        end
        
        function reset(GW)
            GW.n = 0;
            GW.nbPts = 0;
        end
        
        function update(GW,y)
            if GW.n == 0
                GW.sigma = 1;
                GW.alpha(1) = 0;
                GW.d(1) = y;
                GW.n = 1;
                GW.nbPts = 1;
                GW.ysqsum = y * y;
            else
                if GW.n < GW.degree +2 
                    nNew =  int32(GW.n + 1);
                else
                    nNew = int32(GW.n + 0);
                end
                
                sigmaNew = sqrt(GW.sigma * GW.sigma + 1);
                c = GW.sigma / sigmaNew;
                s = 1 / sigmaNew;
                GW.sigma = sigmaNew;
                theta1 = c * s * (GW.nbPts - GW.alpha(1));
                
                if GW.degree > 0
                    theta2 = -s * GW.beta(1);
                else
                    theta2 = 0;
                end
               
                GW.alpha(nNew) = c * c * GW.nbPts + s * s * GW.alpha(1);
                GW.alpha(1) = c * c * GW.alpha(1) + s * s * GW.nbPts;
                GW.beta(1) = c * GW.beta(1);
                GW.d(nNew) = c * y - s * GW.d(1);
                GW.d(1) = c * GW.d(1) + s * y;
                
                for i=1:nNew-2
                    if GW.beta(i) * GW.beta(i) + theta1 * theta1 == 0
                        GW.n=i;
                        return;
                    end
                    
                    zeta = sqrt(GW.beta(i) * GW.beta(i) + theta1 * theta1);
                    c = GW.beta(i) / zeta;
                    s = theta1 / zeta;
                    
                    GW.beta(i) = zeta;
                    t = c * c * GW.alpha(i+1) + 2 * c * s * theta2 + s * s * GW.alpha(nNew);
                    r = c * c * GW.alpha(nNew) - 2 * c * s * theta2 + s * s * GW.alpha(i+1);
                    theta1 = c * s * (GW.alpha(nNew) - GW.alpha(i+1)) + theta2 * (c * c - s * s);
                    theta2 = -s * GW.beta(i+1);
                    GW.alpha(i+1) = t;
                    GW.alpha(nNew) = r;
                    GW.beta(i+1) = c * GW.beta(i+1);
                    t = GW.d(i+1);
                    GW.d(i+1) = c * t + s * GW.d(nNew);
                    GW.d(nNew) = c * GW.d(nNew) - s * t;
                end
                
                if(theta1 ~= 0)
                   GW.beta(nNew-1) = abs(theta1);
                   GW.d(nNew) = sign(theta1) * GW.d(nNew);
                   GW.n = nNew;
                else
                    error('Warning: theta1 = 0');
                end
                
                GW.nbPts = GW.nbPts + 1;
                GW.ysqsum = GW.ysqsum + y * y;
            end
        end
        
        function polyVal = eval(GW,x)
            if GW.n <= GW.degree
               error('Unable to evaluate polynomial for less than degree + 1 points.')
            end
            
            y = zeros(length(GW.d));
            
            if GW.degree == 0
               y(1) = GW.d(1); 
            elseif GW.degree == 1;
               y(1) = GW.d(1) + GW.d(2) / GW.beta(1) * (x - GW.alpha(1));
            else
               y(GW.degree+1) = GW.d(GW.degree+1) / GW.beta(GW.degree); 
               y(GW.degree) = (GW.d(GW.degree) + y(GW.degree+1) * (x - GW.alpha(GW.degree))) / GW.beta(GW.degree-1);
               
               for i= GW.degree -2:-1:1
                  y(i+1) = ( GW.d(i+1) + y(i+2) ...
                             * (x - GW.alpha(i+1)) - y(i+3) ...
                             * GW.beta(i+1)) ...
                             / GW.beta(i);
               end
               
               y(1) = GW.d(1) + y(2) * (x - GW.alpha(1)) - y(3) * GW.beta(1);
            end
           
            polyVal = y(1) / GW.sigma;
        end
        
        function nbPts = getNumSamples(GW)
            nbPts = GW.nbPts;
        end
        
        function c = getOrthogonalCoefficients(GW)
            if GW.n <= GW.degree
               error('Unable to evaluate polynomial for less than degree + 1 points.')
            end
            
            c = zeros(GW.degree + 1,1);
            nf = 1 / GW.sigma;
            
            c(1) = nf * GW.d(1);
            
            for i=1:GW.degree
               nf = nf / GW.beta(i); 
               c(i+1) = GW.d(i+1) * nf;
            end
        end
        
        function approxError = getApproximationError(GW)
            
            if GW.n <= GW.degree
               error('Unable to evaluate approximation error for less than degree + 1 points.')
            end
            
            dsqsum = 0;
            
            for i=0:GW.degree
               dsqsum = dsqsum + GW.d(i+1) * GW.d(i+1); 
            end
            
            if(GW.ysqsum - dsqsum < -1e-8)
                error('Warning: Approximation error negative!')
            end
            
            approxError = max(0,GW.ysqsum - dsqsum);
        end
        
        function exPos = getExtremumPosition(GW)
           
            if GW.degree < 2 ||  GW.n <= GW.degree
                error('Unable to calculate the extremum of the current solution (degree must be >= 2, solution size must be > degree)')
            end
            
            if GW.d(3) == 0
               error('Unable to calculate the extremum of the current solution (curve is zero)')
            end
            
            exPos = int32( round( (GW.alpha(1) + GW.alpha(2)) / 2 - GW.d(2) * GW.beta(2) / (2*GW.d(3)) ) );
        end
        
    end %methods
end %class