% /**
%  * 	This class provides algorithms for up- and downdating least squares
%  * 	polynomial fits to discrete data with orthogonal polynomials using
%  * 	(hyper-)rotations and for evaluating the resulting polynomials.
%  * 
%  *  @see   Fuchs, E.; Gruber, T.; Nitschke, J.; Sick, B.,
%  *			"Online Segmentation of Time Series Based on Polynomial Least-Squares Approximations,"
%  * 			Pattern Analysis and Machine Intelligence, IEEE Transactions on ,
%  * 			vol.32, no.12, pp.2232,2245, Dec. 2010
%  *
%  *  @date   16.08.2013
%  *  @author Dr. Thiemo Gruber, Andr Gensler
%  *
%  *  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES\n
%  *  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF\n
%  *  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR\n
%  *  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES\n
%  *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN\n
%  *  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF\n
%  *  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.\n
%  *
%  *  @copyright Intelligent Embedded Systems, University of Kassel, 2013. All rights reserved.
%  *  @license This library is released under <a href="http://www.gnu.org/licenses/lgpl-3.0.de.html">LGPLv3</a>.
%  */

classdef OrthogonalPolynomialSlidingWindow < handle & matlab.mixin.Copyable
    %SlidingWindow This is the SlidingWindow implementation in Matlab.
    %   For detailed documentation see http://ies-research.de/Software
    
    properties
        degree
        windowSize
        qnorm
        delta
        alpha
        updateCoefs
        values
        valueIdx
        sqsum
    end
    
    methods
        function SW = OrthogonalPolynomialSlidingWindow(degree,windowSize)
            
            if degree < 0
               error('Degree must not be negative') 
            end
            
            if degree >= windowSize
               error('Window size must be greater than degree') 
            end
            
            SW.degree = degree;
            SW.windowSize = windowSize;
            
            SW.init();
            
            m = windowSize -1;
            
            delta = zeros(degree+2,1);
            qnorm = zeros(degree+2,1);
            
            delta(1) = m+1;
            qnorm(1) = delta(1);
            
            qmpo = (m+1) * (m+1);
            
            for i=1:degree+1
                qi = i*i;
                delta(i+1) = qi * (qmpo - qi) / (16 * qi - 4);
                qnorm(i+1) = qnorm(i) * delta(i+1);
            end
            
            SW.delta = zeros(degree +1,1);
            SW.delta(1:degree+1,1) = delta(1:degree+1,1);
            
            SW.qnorm = zeros(degree +1,1);
            SW.qnorm(1:degree+1,1) = qnorm(1:degree+1,1);
            
            gamma = cell(degree+1,1);
            
            gamma{1} = [1; 0; 0];
            
            if degree > 0
               gamma{2} = [1; 1; 0; 0];
               
               for i=2:length(gamma)-1
                  gamma{i+1} = zeros(i+3,1);
                  
                  for j=0:i
                  
                      gamma{i+1}(j+1) = gamma{i}(j+2) * delta(j+2) ...
                          + gamma{i}(j+1) - gamma{i-1}(j+1) * delta(i);
                      
                      if j>0
                         gamma{i+1}(j+1) =  gamma{i+1}(j+1) + gamma{i}(j);
                      end     
                  end
               end 
            end
            
            SW.updateCoefs = cell(degree+1,1);
            SW.updateCoefs{1} = zeros(2,1);
            SW.updateCoefs{1}(1) = 1;
            SW.updateCoefs{1}(2) = -1;
            
            if degree >0
               for i=1:length(SW.updateCoefs)-1
                  
                   SW.updateCoefs{i+1} = zeros(i+2,1);
                   
                   SW.updateCoefs{i+1}(1) = SW.updateCoefs{i}(1) * i / (4.0 *i -2) * (m+ i + 1);
                   SW.updateCoefs{i+1}(2) = SW.updateCoefs{i}(2) *-i / (4.0 *i -2) * (m- i + 1);
                   
                   for j=2:i+1
                      SW.updateCoefs{i+1}(j+1) = -gamma{i+1}(j-1);
                   end
               end
            end
        end
        
        function init(SW)
            SW.alpha = zeros(SW.degree+1,1);
            SW.values = zeros(SW.windowSize,1);
            SW.valueIdx = 0;
            SW.sqsum = 0;
        end
        
        function restart(SW)
           
            valuesOld = SW.values;
            valueIdxOld = SW.valueIdx;
            
            SW.init();
            
            for i=0:SW.windowSize
               SW.update(valuesOld( mod((valueIdxOld + i),length(valuesOld))+1 ) );
            end
        end
        
        function update(SW,y)
            
            for i=0:SW.degree
               alphaNew = SW.alpha(i+1) + y * SW.updateCoefs{i+1}(1) ...
                   + SW.values(SW.valueIdx+1) * SW.updateCoefs{i+1}(2);
               
               for j=0:i-1
                  alphaNew = alphaNew + SW.updateCoefs{i+1}(j+3) * SW.alpha(j+1); 
               end
               
               SW.alpha(i+1) = alphaNew;
            end
            
            SW.sqsum = SW.sqsum + y*y - SW.values(SW.valueIdx+1) * SW.values(SW.valueIdx+1);
            SW.values(SW.valueIdx+1) = y;
            SW.valueIdx = SW.valueIdx+1;
            
            if SW.valueIdx >= length(SW.values)
               SW.valueIdx = 0; 
            end
        end
        
        function val = eval(SW,x)
           
            q = zeros(SW.degree+1);
            q(SW.degree+1) = SW.alpha(SW.degree+1) / SW.qnorm(SW.degree+1);
            
            if SW.degree > 0
               
                x = x - (SW.windowSize -1) / 2.0;
                
                q(SW.degree) = SW.alpha(SW.degree) / SW.qnorm(SW.degree) + x * q(SW.degree+1);
                
                for i=SW.degree-2:-1:0 %!
                    q(i+1) = SW.alpha(i+1) / SW.qnorm(i+1) + x * q(i+2) - SW.delta(i+2) * q(i+3);
                end
            end
            
            val = q(1);
        end
        
        function val = getOrthogonalCoefficients(SW)
           
            coefs = zeros(SW.degree+1,1);
            
            for i=0:length(coefs)-1
               coefs(i+1) = SW.alpha(i+1)  / SW.qnorm(i+1);
            end
            
            val = coefs;
        end
        
        function val = getApproximationError(SW)
           
            tmpSum = SW.sqsum;
            
            for i=0:SW.degree
                tmpSum = tmpSum - SW.alpha(i+1) * SW.alpha(i+1) / SW.qnorm(i+1);
            end
            
            if tmpSum < -1e-8
                error('Warning: Approximation error negative!')
            end
            
            val = max(0,tmpSum);
        end
        
        function val = getExtremumPosition(SW)
           
            if SW.degree <2
                error('Unable to calculate the extremum of the current solution (degree must be >= 2)')
            end
            
            if SW.alpha(3) == 0
                error('Unable to calculate the extremum of the current solution (curve is zero)')
            end
            
            val = int32( round((SW.windowSize -1) / 2.0 - (SW.alpha(2) * SW.qnorm(3)) ...
                / (2.0 * SW.alpha(3) * SW.qnorm(2)) ) );
        end
        
    end

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
end

