classdef TrigonometricPolynomialSlidingWindow < handle & matlab.mixin.Copyable
    
    properties
        degree
        windowSize
        cRe
        cIm
        cosPhi
        sinPhi
        values
        valueIdx
        sqsum
    end
    
    methods
       
        function TW = TrigonometricPolynomialSlidingWindow(degree, windowSize)
            
            if degree < 0
               error('Degree must not be negative') 
            end
            
            if windowSize < 1 || windowSize < degree * 2
                error('Window size must be greater than or equal to degree * 2 and must be at least 1.')
            end
            
            TW.degree = degree;
            TW.windowSize = windowSize;
            
            TW.cosPhi = zeros(degree+1, 1);
            TW.sinPhi = zeros(degree+1, 1);
            
            for k = 0:degree
               phi = 2 * pi * k / TW.windowSize;
               
               TW.cosPhi(k+1) = cos(phi);
               TW.sinPhi(k+1) = sin(phi);
            end
            
            TW.init();
            
        end
        
        function init(TW)
            
            TW.cRe = zeros(TW.degree+1, 1);
            TW.cIm = zeros(TW.degree+1, 1);
            TW.values = zeros(TW.windowSize, 1);
            TW.valueIdx = 1;
            TW.sqsum = 0;
            
        end
        
        function reset(TW)
           
            TW.init();
            
        end
        
        function update(TW, y)
           
            for k = 1:length(TW.cRe)
                
               s = TW.cRe(k) + y - TW.values(TW.valueIdx);
               TW.cRe(k) = TW.cosPhi(k) * s - TW.sinPhi(k) * TW.cIm(k);
               TW.cIm(k) = TW.cosPhi(k) * TW.cIm(k) + TW.sinPhi(k) * s;
               
            end
            
            TW.sqsum = TW.sqsum + y * y - TW.values(TW.valueIdx) * TW.values(TW.valueIdx);
            
            TW.values(TW.valueIdx) = y;
            
            TW.valueIdx = TW.valueIdx + 1;
            
            if TW.valueIdx > length(TW.values)
               TW.valueIdx = 1;
            end
        end
        
        function val = eval(TW, x)
           
            p = double(TW.cRe(1));
            
            cond = TW.degree - 2 * TW.degree / TW.windowSize;
            
            for k = 1:cond+1
               
                komegax = k * 2 * pi / length(TW.values) * x;
               
                p = p + 2 * TW.cRe(k+1) * cos(komegax) - 2 * TW.cIm(k+1) * sin(komegax);
                
            end
            
            if TW.windowSize == 2 * TW.degree
               
                p = p + TW.cRe(length(TW.cRe)) * cos(pi * x);
                
            end
            
            val = p / length(TW.values);
            
        end
        
        function val = getCoefficients(TW)
            
           val = [TW.cRe;TW.cIm];
            
        end
        
        function val = getApproximationError(TW)
            
            qnorms = zeros(2,TW.degree+1);
            
            for x = 1:TW.windowSize
                for k = 1:length(TW.cRe)
                    
                    komegax = (k-1) * 2 * pi / TW.windowSize * x;
                    qnorms(1,k) = qnorms(1,k) + cos(komegax).^2;
                    qnorms(2,k) = qnorms(2,k) + cos(komegax).^2;
                    
                end
            end
            
            tmpSum = TW.sqsum;
            
            for i = 0:TW.degree
                tmpSum = tmpSum - TW.cRe(i+1).^2 /qnorms(1,i+1);
            end

            cond = TW.degree - 2 * TW.degree / TW.windowSize;
            
            for i = 1:cond+1
               tmpSum = tmpSum - TW.cIm(i+1).^2 / qnorms(2,i+1); 
                
            end
            
            if tmpSum < -1e-8
                
                error('Warning: Approximation error negative!')
                
            end
            
            val = max(0, tmpSum);
        end
    end
end