/**
 *  @file   TrigonometricPolynomialSlidingWindow.cpp
 *  @brief  TrigonometricPolynomialSlidingWindow class implementation
 *  @date   26.03.2013
 *  @author Dr. Thiemo Gruber, Peter Rudolph
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES\n
 *  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF\n
 *  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR\n
 *  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES\n
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN\n
 *  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF\n
 *  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.\n
 *
 *  @copyright Intelligent Embedded Systems, University of Kassel, 2013. All rights reserved.
 *  @license This library is released under <a href="http://www.gnu.org/licenses/lgpl-3.0.de.html">LGPLv3</a>.
 */
#include "TrigonometricPolynomialSlidingWindow.hpp"

#include <iostream>
#include <cmath>

TrigonometricPolynomialSlidingWindow::TrigonometricPolynomialSlidingWindow(int degree, int windowSize)
{
  if (degree < 0)
      throw ("Degree must not be negative.");

  if (degree >= windowSize)
      throw ("Window size must be greater than degree.");

  this->degree = degree;
  this->windowSize = windowSize;

  for(size_t i=0;i<degree +1;++i)
    this->cosPhi.push_back(double());
  for(size_t i=0;i<degree +1;++i)
    this->sinPhi.push_back(double());

  for (int k = 0; k <= degree; k++) {
    double phi = 2 * M_PI * k / windowSize;
    cosPhi[k] = cos(phi);
    sinPhi[k] = sin(phi);
  }
  init();

}

void TrigonometricPolynomialSlidingWindow::init()
{
  this->cRe.clear();
  this->cIm.clear();
  this->values.clear();

  for(size_t i=0;i<degree +1;++i)
    this->cRe.push_back(double());
  for(size_t i=0;i<degree +1;++i)
    this->cIm.push_back(double());

  for(size_t i=0;i<windowSize;++i)
    this->values.push_back(double());

  this->valueIdx = 0;
  this->sqsum = 0;
}

void TrigonometricPolynomialSlidingWindow::restart()
{
  init();
}

void TrigonometricPolynomialSlidingWindow::update(double y)
{
  for (int k = 0; k < cRe.size(); k++) 
  {
    double s = cRe[k] + y - values[valueIdx];
    cRe[k] = cosPhi[k] * s - sinPhi[k] * cIm[k];
    cIm[k] = cosPhi[k] * cIm[k] + sinPhi[k] * s;
  }

  sqsum += y * y - values[valueIdx] * values[valueIdx];

  values[valueIdx++] = y;

  if (valueIdx >= values.size())
      valueIdx = 0;
}

double TrigonometricPolynomialSlidingWindow::eval(double x)
{
  double p = cRe[0];

  for (int k = 1; k <= degree - 2 * degree / windowSize; k++) 
  {
      double komegax = k * 2 * M_PI / values.size() * x;
      p += 2 * cRe[k] * cos(komegax) - 2 * cIm[k] * sin(komegax);
  }

  if (2 * degree == windowSize) 
  {
      p += cRe[cRe.size() - 1] * cos(M_PI * x);
  }

  return p / values.size();
}

std::vector< std::vector<double> > TrigonometricPolynomialSlidingWindow::getCoefficients()
{
  std::vector< std::vector<double> > c;
  c.push_back(cRe);
  c.push_back(cIm);
  return c;
}

double TrigonometricPolynomialSlidingWindow::getApproximationError()
{
  std::vector< std::vector<double> > qnorms;
  std::vector<double> qnorm_cos(degree+1);
  std::vector<double> qnorm_sin(degree+1);

  qnorms.push_back(qnorm_cos);
  qnorms.push_back(qnorm_sin);

  for (int x = 0; x < windowSize; x++) {
      for (int k = 0; k < cRe.size(); k++) {
          double komegax = k * 2 * M_PI / windowSize * x;
          qnorms[0][k] += pow(cos(komegax), 2);
          qnorms[1][k] += pow(sin(komegax), 2);
      }
  }

  double tmpSum = sqsum;

  for (int i = 0; i <= degree; i++) {
      tmpSum -= pow(cRe[i], 2) / qnorms[0][i];
  }

  for (int i = 1; i <= degree - 2 * degree / windowSize; i++) {
      tmpSum -= pow(cIm[i], 2) / qnorms[1][i];
  }

  if (tmpSum < -1e-8)
  {
    std::cerr << "Warning: Approximation error negative!" << tmpSum << std::endl;
    return 0;
  }
  return tmpSum;
}
